from typing import Union, List

from pypleasant.artifacts import Database, Folder, Entry, Attachment


class PleasantElementNotFound(Exception):
    def __init__(self, complete_path: str, entry: str):
        super().__init__(f"Could not find {entry} in {complete_path}")


class PleasantEntryNotDistinct(Exception):
    def __init__(self, complete_path: str, entry: str):
        super().__init__(f"{entry} from {complete_path} exists in custom fields and attachments.")


class PathParser:
    def __init__(self, database: Database):
        self.database = database
        self._complete_path = None

    def lookup(self, path_as_str: str) -> Union[Folder, Entry, Attachment, str]:
        self._complete_path = path_as_str
        path = path_as_str.strip("/").split("/")
        return self.database if path == [''] else self._traverse_path(path, self.database)

    def _traverse_path(self, path: List[str], parent: Union[Database, Folder, Entry]) -> Union[
        Folder, Entry, Attachment, str]:
        if len(path) == 0:
            return parent
        elif isinstance(parent, Entry) and len(path) == 1:
            return self._get_entry_data(parent, path[0])
        else:
            child_name = path[0]
            if child_name not in parent:
                raise PleasantElementNotFound(self._complete_path, child_name)

            return self._traverse_path(path[1:], parent[child_name])

    def _get_entry_data(self, entry: Entry, data_name: str) -> Union[Attachment, str]:
        in_attachments = data_name in entry.attachments
        in_custom_fields = data_name in entry.custom_fields

        if in_attachments and in_custom_fields:
            raise PleasantEntryNotDistinct(self._complete_path, data_name)
        elif in_attachments:
            return entry.attachments[data_name]
        elif data_name in entry.custom_fields:
            return entry.custom_fields[data_name]
        else:
            raise PleasantElementNotFound(self._complete_path, data_name)



import base64
from datetime import datetime, timedelta

import requests


class BadCredentials(Exception):
    def __init__(self):
        super().__init__("Invalid username or password")


class PleasantAPIToken:
    def __init__(self, token_as_json):
        self.access_token = token_as_json["access_token"]
        self._expire_date = datetime.now() + timedelta(seconds=token_as_json["expires_in"])

    @property
    def expired(self):
        return datetime.now() > self._expire_date


class PleasantAPI:
    def __init__(self, url: str, user: str, password: str, verify_https: bool = True):
        self.base_url = url
        self.user = user
        self.password = password
        self.verify_https = verify_https
        self._token = None

        if not self.verify_https:
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    @property
    def token(self) -> str:
        if self._token is None or self._token.expired:
            data = {
                "grant_type": "password",
                "username": self.user,
                "password": self.password
            }

            response = requests.get(f'{self.base_url}/oauth2/token', data=data, verify=self.verify_https)
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as e:
                if response.status_code == 400 and response.json()["error"] == "invalid_grant":
                    raise BadCredentials
                else:
                    raise e
            self._token = PleasantAPIToken(response.json())

        return self._token.access_token

    def _request(self, url: str) -> requests.Response:
        headers = {
            'Authorization': f'Bearer {self.token}',
            'Content-Type': 'application/json',
            'Cache-Control': 'no-cache'
        }

        response = requests.get(url, headers=headers, verify=self.verify_https)
        response.raise_for_status()
        return response

    def get_credential(self, entry_id: str) -> str:
        url = f'{self.base_url}/api/v5/rest/entries/{entry_id}/password'
        return self._request(url).text[1:-1]

    def get_attachment(self, entry_id: str, attachment_id: str) -> bytes:
        url = f"{self.base_url}/api/v5/rest/entries/{entry_id}/attachments/{attachment_id}"
        decoded_file = self._request(url).json()["FileData"]
        return base64.b64decode(decoded_file)

    def get_db(self) -> dict:
        url = f"{self.base_url}/api/v5/rest/folders"
        db_hierarchy_as_json = self._request(url).json()
        return db_hierarchy_as_json



import base64
import pathlib
from collections import UserDict

from pypleasant.api import PleasantAPI


class Attachment:
    def __init__(self, attachment_as_json: dict, api: PleasantAPI):
        self.api = api
        self._entry_id = attachment_as_json["CredentialObjectId"]
        self._attachment_id = attachment_as_json["AttachmentId"]
        self.name = attachment_as_json["FileName"]

    @property
    def data(self) -> bytes:
        return self.api.get_attachment(self._entry_id, self._attachment_id)

    def __str__(self):
        return base64.b64encode(self.data).decode()

    def download(self, output_file_path: pathlib.Path = None):
        output_file_path = output_file_path or pathlib.Path(f"./{self.name}")
        output_file_path.write_bytes(self.data)


class Attachments(UserDict):
    def download(self, output_dir: pathlib.Path = None):
        output_dir = output_dir or pathlib.Path(f"./pleasant_attachments")
        if not output_dir.exists():
            output_dir.mkdir()

        for file_name, attachment in self.data.items():
            attachment.download(output_dir / file_name)


class Entry:
    def __init__(self, entry_as_json, api: PleasantAPI):
        self.api = api
        self._entry_id = entry_as_json["Id"]
        self.name = entry_as_json["Name"]
        self.username = entry_as_json["Username"]
        self.url = entry_as_json["Url"]
        self.custom_fields = entry_as_json["CustomUserFields"]

        attachments_as_dict = {}
        for attachment_as_json in entry_as_json["Attachments"]:
            attachments_as_dict[attachment_as_json["FileName"]] = Attachment(attachment_as_json, api)
        self.attachments = Attachments(attachments_as_dict)

    @property
    def password(self) -> str:
        return self.api.get_credential(self._entry_id)


class Folder(UserDict):
    def __init__(self, folder_as_json: dict, api: PleasantAPI):
        self.name = folder_as_json["Name"]
        entries = {entry_as_json["Name"]: Entry(entry_as_json, api) for entry_as_json in
                   folder_as_json["Credentials"]}
        folders = {folders_as_json["Name"]: Folder(folders_as_json, api) for folders_as_json in
                   folder_as_json["Children"]}
        super().__init__({**entries, **folders})


class Database(Folder):
    def __init__(self, api: PleasantAPI):
        super().__init__(api.get_db(), api)



import argparse
import logging
import os
import pathlib
import sys
from getpass import getpass
from typing import Tuple, List

import requests

from pypleasant.api import PleasantAPI, BadCredentials
from pypleasant.artifacts import Database, Entry
from pypleasant.pathparser import PathParser, PleasantElementNotFound


def parse_cmd() -> Tuple[str, str, str, List[str], pathlib.Path, str, str, str, bool]:
    parser = argparse.ArgumentParser("pleasant-client")
    attribute_group = parser.add_mutually_exclusive_group(required=True)
    attribute_group.add_argument("--username", action="store_true", help="print the username")
    attribute_group.add_argument("--password", action="store_true", help="print the password")
    attribute_group.add_argument("--url", action="store_true", help="print the URL")
    attribute_group.add_argument("--custom-field", type=str, nargs='?',
                                 help="print the given custom field")
    attribute_group.add_argument("--attachments", type=str, nargs='*',
                                 help="download the given attachment(s); if no attachment is given, all attachments are downloaded")

    parser.add_argument("--download-dir", type=str,
                        help="attachments are downloaded to this directory (DEFAULT: '.', env var: PYPLEASANT_DOWNLOAD_DIR)")

    parser.add_argument("--api-url", type=str, help="URL of the pleasant server API (env var: PYPLEASANT_API_URL)")
    parser.add_argument("--api-user", type=str, help="user for the pleasant server API (env var: PYPLEASANT_API_USER)")
    parser.add_argument("--api-password", type=str,
                        help="password for the pleasant server API (env var: PYPLEASANT_API_PASSWORD)")
    parser.add_argument("--disable-cert-check", action="store_true",
                        help="disable HTTPS cert check (env var: PYPLEASANT_DISABLE_CERT_CHECK)",
                        dest="disable_cert_check")
    parser.add_argument("--verbose", action="store_true", help="activate verbose output")
    parser.add_argument("--debug", action="store_true", help="activate debug output (env var: PYPLEASANT_DEBUG)")

    parser.add_argument("path", type=str,
                        help="the path on the pleasant server to the credential entry, e.g. /Development/git (env var: PYPLEASANT_PATH_TO_ENTRY)")
    args = parser.parse_args()

    if args.username:
        entry_attribute = "username"
    elif args.password:
        entry_attribute = "password"
    elif args.url:
        entry_attribute = "url"
    elif args.custom_field:
        entry_attribute = "custom_field"
    else:
        entry_attribute = "attachments"

    custom_field_key = args.custom_field
    attachment_file_names = args.attachments
    download_dir = args.download_dir or os.getenv("PYPLEASANT_DOWNLOAD_DIR") or '.'

    api_url = args.api_url or os.getenv("PYPLEASANT_API_URL") or input("Pleasant Server URL: ")
    api_user = args.api_user or os.getenv("PYPLEASANT_API_USER") or input("Pleasant User: ")
    api_password = args.api_password or os.getenv("PYPLEASANT_API_PASSWORD") or getpass()
    disable_cert_check = args.disable_cert_check or os.getenv("PYPLEASANT_DISABLE_CERT_CHECK", "").lower() == "true"
    verify_https = not disable_cert_check

    activate_debug = args.debug or os.getenv("PYPLEASANT_DEBUG")
    if activate_debug:
        logging.root.setLevel(logging.DEBUG)
    elif args.verbose:
        logging.root.setLevel(logging.INFO)

    path = args.path or os.getenv("PYPLEASANT_PATH_TO_ENTRY")

    return path, entry_attribute, custom_field_key, attachment_file_names, pathlib.Path(
        download_dir), api_url, api_user, api_password, verify_https


def print_prettified_exception(exception: BaseException) -> None:
    prettify_output = logging.root.level == logging.DEBUG
    if not prettify_output:
        raise exception
    else:
        standard_error_message = "Internal error. For more details activate debug output with --debug"
        logging.error(exception.args[0] if exception.args else standard_error_message)


class NotAPleasantEntry(Exception):
    def __init__(self, path: str):
        super().__init__(f"{path} is not an entry")


class PleasantAPIConnectionError(Exception):
    def __init__(self, url: str):
        super().__init__(f"Could not connect to {url}")


def lookup(url: str, user: str, password: str, verify_https: bool, path: str) -> Entry:
    api = PleasantAPI(url, user, password, verify_https)
    database = Database(api)
    entry = PathParser(database).lookup(path)

    if not isinstance(entry, Entry):
        raise NotAPleasantEntry(path)
    else:
        return entry


def main() -> None:
    logging.basicConfig(stream=sys.stderr,
                        level=logging.DEBUG if os.getenv("PYPLEASANT_DEBUG") else None,
                        format="%(levelname)s: %(message)s")

    try:
        path, entry_attribute, custom_field_key, attachment_file_names, download_dir, api_url, api_user, api_password, verify_https = parse_cmd()

        entry = lookup(api_url, api_user, api_password, verify_https, path)
        if entry_attribute == "username":
            print(entry.username)
        elif entry_attribute == "password":
            print(entry.password)
        elif entry_attribute == "url":
            print(entry.url)
        elif entry_attribute == "custom_field":
            print(entry.custom_fields[custom_field_key])
        elif entry_attribute == "attachments":
            attachment_file_names = attachment_file_names if attachment_file_names else [file_name for file_name in
                                                                                         entry.attachments]
            for file_name in attachment_file_names:
                print(f"{path}: {file_name} -> {download_dir / file_name}")
                entry.attachments[file_name].download(download_dir / file_name)
    except BadCredentials as e:
        print_prettified_exception(e)
    except NotAPleasantEntry as e:
        print_prettified_exception(e)
    except PleasantElementNotFound as e:
        print_prettified_exception(e)
    except requests.exceptions.ConnectionError as e:
        print_prettified_exception(PleasantAPIConnectionError(e.request.url))
    except KeyboardInterrupt:
        pass
    except Exception as e:
        print_prettified_exception(e)



import pathlib

from setuptools import setup, find_packages

VERSION = "1.0.2"
PROJECT_DIR = pathlib.Path(__file__).parent
SRC_DIR = PROJECT_DIR / "src"

setup(
    name='pypleasant',
    version=VERSION,
    description="Python library and script to access the API of the pleasant password server.",
    long_description=(PROJECT_DIR / "README.md").read_text(),
    long_description_content_type="text/markdown",
    license="MIT",
    author='Dominik Bermühler',
    author_email='dominik.bermuehler@googlemail.com',
    url='https://github.com/dbermuehler/pypleasant',
    install_requires=['requests'],
    package_dir={'': str(SRC_DIR)},
    packages=find_packages(str(SRC_DIR)),
    entry_points={'console_scripts': ['pleasant-cli=pypleasant.cli:main']},
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Topic :: Security",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License"
    ],
    python_requires='>=3.6'
)
